#!/usr/bin/env python 
import sys
import os
import yaml

from .utils  import system_output
from .logging import error, info, fatal

RESOURCES = 'resources.yaml'

def main():
    if len(sys.argv) == 1:
        dirname = '.'
    else:
        dirname = sys.argv[1]

    output = os.path.join(dirname, RESOURCES)
    if os.path.exists(output):
        raise Exception('Output file %s already exist.' % output)
    
    cols = 80
    
    def format_string(s):
        return s.ljust(cols)[:cols]
    
    repolist = []
    
    def append(r):
        if 'sub' in r:
            info(format_string("Found sub in: %s" % repo))
        elif 'dir' in r:
            info(format_string('Found git in: %s' % repo))
        repolist.append(r)
        
    def consider(repo):
        if RESOURCES in repo:
            append({'sub': os.path.relpath(repo, dirname)})
            
        if '.git' in repo:
            git_dir = repo
            destination = os.path.relpath(os.path.dirname(git_dir), dirname)
            found = {'dir': destination}
            info('Found git in: %s' % found['dir'])
                
            try:
                url, branch = get_url_branch(git_dir)
                if branch != 'master':
                    found['branch'] = branch
                found['url'] = url
                if not 'git' in found['url']:
                    found['type'] = 'git'
            except Exception as e:
                error('Could not locate url for %r: %s' % (git_dir, e))
                
            append(found)
    
    
    def mark():
        mark.count += 1
        markers = ['-', '/', '|', '\\']
        return markers[mark.count % len(markers)]
      
    mark.count = 0
    
    def log(s):
        sys.stderr.write(format_string('%s %5d %s' % (mark(), mark.count, s)))
        sys.stderr.write('\r')
   
    for repo in find_repos(dirname, log=log, shallow=True):
        consider(repo)
   
    if len(repolist) == 0:
        fatal('No repos found in %r.' % dirname)
 
    with open(output, 'w') as f:
        for repo in repolist:
            f.write('---\n')
            yaml.dump(repo, f, default_flow_style=False)
 

def get_url_branch(git_dir):
    repo = os.path.dirname(git_dir)
    repo = os.path.realpath(repo)

    remotes = system_output(repo, 'git remote')
    if not 'origin' in remotes.split():
        raise Exception("No remote origin found for %r." % repo)
            
    try:
        result = system_output(repo, 'git remote show origin')
    except Exception as e:
        raise Exception("Could not get origin URL for %r: %s" % (repo, e))
                
    tokens = result.split()
    # XXX add checks
    url = tokens[tokens.index('URL:') + 1]
    
    result = system_output(repo, 'git branch --color=never')
    for line in result.split('\n'):
        if line[0] == '*':
            branch = line.split()[1]
            break
    else:
        print("Could not parse branch from %s, using master." % 
              result.__repr__())
        branch = 'master'

    return url, branch
    

def find_repos(directory, log=lambda _: None, followlinks=False, shallow=True):
    reasonable = shallow
    resources = os.path.join(directory, RESOURCES)
    if os.path.exists(resources):
        yield resources
        if reasonable: 
            return

    git_repo = os.path.join(directory, '.git')
    if os.path.exists(git_repo):
        yield git_repo
        if reasonable: 
            return
        
    for root, dirs, _ in os.walk(directory, followlinks=followlinks):
        log(root)
        
        for dirname in list(dirs):
            resources = os.path.join(root, dirname, RESOURCES)
            if os.path.exists(resources):
                yield resources
                dirs.remove(dirname)
            else:
                git_repo = os.path.join(root, dirname, '.git')
                if os.path.exists(git_repo):
                    yield git_repo
                    if reasonable: 
                        dirs.remove(dirname)
                    
if __name__ == '__main__':
    main()





class PatienceException(Exception):
    pass

class ActionException(PatienceException):
    pass

class ConfigException(PatienceException):
    def __init__(self, error, config):
        self.error = error
        self.config = config
        
    def __str__(self):
        return "%s\n%s" % (self.error, self.config)


class UserError(PatienceException):
    pass



from . import checkout
from . import develop
from . import fetch
from . import install
from . import merge
from . import push
from . import status

from . import unittests

from . import branches
 
from . import status_branches



from patience.action import Action
from patience.structures import ActionException

__all__ = ['Develop']


class Develop(Action):
    
    def __init__(self): 
        Action.__init__(self, parallel=False, any_order=False)

    def single_action_starting(self, resource):  # @UnusedVariable
        print 'ciao' 
        pass

    def single_action_result_display(self, resource, result): 
        pass
    
    def single_action(self, r):
        if r.is_downloaded():
            r.develop()
        else:
            raise ActionException('Resource %s not ready.' % r)
    
Action.actions['develop'] = Develop()



from patience.action import Action
from patience.structures import ActionException


__all__ = ['Install']


class Install(Action):
    
    def __init__(self): 
        Action.__init__(self, parallel=False, any_order=False)

    def single_action_starting(self, resource): 
        pass

    def single_action_result_display(self, resource, result): 
        pass
    
    def single_action(self, r):
        if r.is_downloaded():
            r.install()
        else:
            raise ActionException('Resource %s not ready.' % r)
    
Action.actions['install'] = Install()



from patience.action import Action
from patience.structures import ActionException

__all__ = ['Checkout']


class Checkout(Action):
    
    def __init__(self): 
        Action.__init__(self, parallel=False, any_order=False)

    def single_action_starting(self, resource): 
        pass

    def single_action_result_display(self, resource, result): 
        pass
    
    def applicable(self, r):
        return not r.is_downloaded()
    
    def single_action(self, r):
        if r.is_downloaded():
            raise ActionException('Already downloaded %s' % r)

        r.checkout()
    
Action.actions['checkout'] = Checkout()




from patience.action import Action
from patience.status_string import status2string, StatusResult, status_fields


__all__ = ['StatusBranches']


class StatusBranches(Action):
    """ Lists the differences with all remote branches """
    
    def __init__(self): 
        Action.__init__(self, parallel=True, any_order=False)

    def single_action_starting(self, resource):  # @UnusedVariable
        return None

    def result_display_start(self):
        return ""
#         """  unknown_       __merges    
# modified  |     |    pushes     _branch status
#     |     |     |       |      |
# """

    def single_action_result_display(self, resource, result):
        return None
        # return str(result)
#         if not isinstance(result, Exception):
#             return status2string(resource, result)
    
    def single_action(self, r):
        return r.list_differences_with_remote_branches()
        

    def summary(self, resources, results):
        s = ""
        from collections import defaultdict
        counter = defaultdict(lambda: 0)
        counter2 = defaultdict(lambda: 0)
        for r, v in results.items():
            for b in v.keys():
                counter[b] += 1
                counter2[r] += 1
            
        branches = sorted(counter.keys(), key=lambda x: (-counter[x]))
        resources = sorted(counter2.keys(), key=lambda x: (-counter2[x]))
        
        table = []
        for r in resources:
            v = results[r]
            if isinstance(v, Exception):
                continue
            row = []
            row.append(str(r))
            for b in branches:
                if not b in v:
                    cell = ''
                else:
                    npush, simple_push, nmerge, simple_merge = v[b]
                    if npush or nmerge:
                        cell = ''
                        if npush:
                            cell += 'P%d' % npush
                            if not simple_push:
                                cell += '!'
                        if nmerge:
                            if npush:
                                cell += ' '
                            cell += 'M%d' % nmerge
                            if not simple_merge:
                                cell += '!'
                    else: 
                        cell = '='        
                row.append(cell)
            table.append(row)
        
        T = [[''] + branches] + table

        s += format_table(T)
        return s
        
def format_table(T, between='  '):
    s = ""
    nrows = len(T)
    ncols = len(T[0])
    column_sizes = [max([len(T[row][col]) for row in range(nrows)]) 
                    for col in range(ncols)] 
    
    for row in T:
        for cell, colsize in zip(row, column_sizes):
            c = cell.ljust(colsize)
            s += c + between
        s += '\n'
    
    return s



Action.actions['status-branches'] = StatusBranches()




from system_cmd import system_run
from .structures import ActionException
import os

def replace_variables(path, rules):
    for k, v in rules:
        if path.startswith(v):
            # print("  applied %s => %s" % (v, k))
            path = path.replace(v, k)
    return path

def get_friendly(path, use_environment=True):  # TODO: make switch
    original = path
    options = []
    
    options.append(os.path.relpath(path, os.getcwd()))

    home = os.path.expanduser('~')
    
    rules = []
    rules.append(('~', home)) 
    
    if use_environment:
        envs = dict(os.environ)
        # remove unwanted 
        for e in list(envs.keys()):
            if 'PWD' in e:
                del envs[e]
    
        for k, v in envs.items():
            if v:
                rules.append(('$%s' % k, v))
        
    # apply longest first
    rules.sort(key=lambda x: (-len(x[1])))
    path = replace_variables(path, rules)
    
    options.append(path)

    weight_doubledot = 5
    
    def score(s):
        # penalize '..' a lot
        s = s.replace('..', '*' * weight_doubledot)
        return len(s)
        
    options.sort(key=score)
    
    if False:
        print('Options for %s' % original)
        for o in options:
            print('- %4d %s' % (score(o), o))
        
    result = options[0]
    
    # print('Converted %s  => %s' % (original, result))

    return result

class Resource(object): 
    def __init__(self, config):
        self.config = config
        self.destination = config['dir']
        
        if not os.path.isabs(self.destination):
            msg = 'Expected all abspath %r' % config
            raise Exception(msg)
        self.short_path = config.get('nick', get_friendly(self.destination))
        
        if os.path.exists(os.path.join(self.destination, 'setup.py')):
            self.config['install'] = 'setuptools'

        if not os.path.exists(self.destination):
            # print('Resource warning: %r does not exist %r' % (self.destination, self.config))
            pass

    def is_downloaded(self):
        return os.path.exists(self.destination)
    
    def __str__(self):
        return self.short_path
    
    def __repr__(self):
        return 'Resource(%r)' % self.destination
    
    def checkout(self):    
        pass

    def update(self):
        pass
        
    def develop(self):
        install_type = self.config.get('install', None)
        if install_type is None:
            self.badconf("No setup method known.")

        if install_type == 'setuptools':
            self.run('python setup.py develop')

    def current_revision(self):
        return None

    def something_to_commit(self):
        return False

    def commit(self):
        pass

    def install(self):
        install_type = self.config.get('install', None)
        if install_type is None:
            self.badconf("No setup method known.")

        if install_type == 'setuptools':
            self.run('python setup.py develop')
        elif install_type == 'cmake':
            # XXX: qui come va?
            system_run(self.destination, 'cmake -DCMAKE_INSTALL_PREFIX=${BVENV_PREFIX} .')
        elif install_type == 'make':
            self.run('make')
            # XXX:
            system_run(self.destination, 'make install')
        
        else:
            raise ActionException('Unknown install type %r.' % install_type)
            
    
    def unittests(self):
        install_type = self.config.get('install', None)
        if install_type is None:
            self.badconf("No setup method known.")

        if install_type == 'setuptools':
            self.run('python setup.py nosetests')
        elif install_type == 'cmake':
            pass
        elif install_type == 'make':
            pass        
        else:
            raise ActionException('Unknown install type %r.' % install_type)
            
            



from .resources import Resource

__all__ = ['Subversion']

class Subversion(Resource):
    def __init__(self, config):
        Resource.__init__(self, config)
        self.url = config['url']

    def checkout(self):
        system_cmd_fail('.', 'svn checkout %s %s' % (self.url, self.destination))

    def update(self):
        system_cmd_fail(self.destination, 'svn update')

    def something_to_commit(self):
        return "" != system_output('svn status %s' % (self.destination))

    def commit(self):
        system_cmd_fail(self.destination, 'svn commit')

    def current_revision(self):
        out = system_output('svnversion %s' % self.destination)
        out = out.split()[0]
        return out




from patience.action import Action
from patience.structures import ActionException


__all__ = ['Install']


class Push(Action):
    
    def __init__(self):
        Action.__init__(self, parallel=True, any_order=True)

    def single_action_started(self, resource): 
        pass
    
    def single_action_result_display(self, resource, result): 
        pass

    def single_action(self, r):
        if not r.is_downloaded():
            raise ActionException('Not downloaded %s' % r)

        # TODO: add branch check
        if r.something_to_push() and r.simple_push():
            r.push()

    def applicable(self, r):
        return r.is_right_branch()
                
Action.actions['push'] = Push()
            



from .structures import ActionException
import sys
from abc import ABCMeta, abstractmethod


__all__ = ['Action']

# from contracts import ContractsMeta


class Action(object):
    
    __metaclass__ = ABCMeta

    def single_action_started(self, resource): 
        pass
    
    def result_display_start(self):
        return None
    
    @abstractmethod
    def single_action_result_display(self, resource, result): 
        pass

    @abstractmethod
    def single_action(self, r):
        pass
    
    def applicable(self, r):  # @UnusedVariable
        """ Return True if it is applicable to this resource. """
        return True
    
    actions = {}
    
    def __init__(self, parallel=True, any_order=False):
        self.parallel = parallel
        self.any_order = any_order
        
    def filter_applicable(self, resources):
        """ Returns the resoruces for which this action is applicable. """
        return filter(self.applicable, resources)
        
    def go(self, resources, force_sequential=False,
                 max_processes=3, stream=sys.stdout,
                 console_status=False,
                 show_operations=False,
                 show_stdout=False):
        
        s = self.result_display_start()
        if s is not None:            
            if stream:
                stream.write(s)
  
        resources = self.filter_applicable(resources)
         
        for r in resources:
            r.show_operations = show_operations
            r.show_stdout = show_stdout 
            
        if not self.parallel or force_sequential:
            return self._go_sequential(resources, stream,
                     console_status=console_status)
        else:
            return self._go_parallel(resources, stream,
                    any_order=self.any_order,
                    max_processes=max_processes,
                    console_status=console_status)
    
    def _go_sequential(self, resources, stream, console_status):
        results = {}
        for i, r in enumerate(resources): 

            if console_status:
                write_console_status("%3d/%d %s" % (i + 1, len(resources), r))

            try:
                result = self.single_action(r)
            except ActionException as e:
                result = e

            if stream:
                m2 = self.single_action_result_display(r, result)
                write_message(stream, r, m2)

            results[r.short_path] = result
            
        if stream:
            m3 = self.summary(resources, results)
            if m3: 
                stream.write('%s\n' % m3)
    
        return results
    

    def _go_parallel(self, resources, stream, any_order, max_processes,
        console_status):

        from multiprocessing import Pool, TimeoutError
        pool = Pool(processes=max_processes)            
        
        handles = {}
        for r in resources: 
            handles[r] = pool.apply_async(slave, [(self, r)])

        results = {}
        r2messages = {}
        to_write = list(resources)
        while handles:
            # print("handles: %d" % len(handles))
            # print(" to write: %d" % len(to_write))
            # print(" r2message: %d" % len(r2messages))
            # print(" results:  %d" % len(results))
            if console_status:
                sys.stderr.write("%3d to go (%3d to write %d)      \r" % 
                (len(handles), len(r2messages), len(to_write)))
            for r, res in list(handles.items()):
                try:
                    results[r] = res.get(timeout=0.05)
                except TimeoutError:
                    continue
                except ActionException as e:
                    results[r] = e
            
                del handles[r]
                    
                m2 = self.single_action_result_display(r, results[r])
                if any_order:
                    write_message(stream, r, m2)
                else:
                    r2messages[r] = m2

            while to_write and to_write[0] in r2messages:
                which = to_write[0]
                r = to_write.pop(to_write.index(which))
                m = r2messages[r]
                del r2messages[r]
                write_message(stream, r, m)
                
        if stream:
            m3 = self.summary(resources, results)
            if m3:
                stream.write('%s\n' % m3)

    def summary(self, resources, results):  # @UnusedVariable
        s = ''
        for path, res in results.items():
            if isinstance(res, Exception):
                e = str(res).split('\n')[0]
                s += '{0:<30}: {1}\n'.format(path, e)
        return s
        
# Load actions 
from . import actions  # @UnusedImport

    


def slave(task):
    action, resource = task
    return action.single_action(resource)
    
    
def write_message(stream, r, m):
    if not m: 
        return
    name = '%s' % r
    if name in m:
        stream.write('%s\n' % m)
    else:
        stream.write('%s: %s\n' % (name, m))
        
def write_console_status(s):
    s = s.ljust(100)  # TODO: add correct lengt
    sys.stderr.write(s + '\r')
    


from .git import Git
from .resources import Resource
from .structures import ConfigException
from .subversion import Subversion
import os
import yaml


def instantiate(config, base_dir='.'):
    """ 
        config['url']
        config['dir']
        config['type'] = git, subversion
    
    """
    
    if not 'url' in config:
        raise ConfigException('Missing key "url".', config)
        
    if not any([x in config for x in ['dir', 'destination']]):
        raise ConfigException('Missing key "dirname".', config)
    
    dirname0 = config.get('dir', config.get('destination'))
    dirname = os.path.expandvars(dirname0)
    dirname = os.path.expanduser(dirname)
    dirname = os.path.join(base_dir, dirname)
    dirname = os.path.abspath(dirname)
    dirname = os.path.realpath(dirname)
    config['dir'] = dirname                 
                                   
#     if not os.path.exists(dirname):
#         print('warn: %r does not exist (%s from %s)' % (dirname, dirname0, base_dir))
#     else:
#         print('found %r %r' % (dirname, config))
                    
    if not 'type' in config:
        url = config['url']
        if 'git' in url:
            config['type'] = 'git'
        elif 'svn' in url:
            # guess git
            config['type'] = 'subversion'
        else:
            raise ConfigException('Could not guess type from url', config)
    
    res_type = config['type']
    if res_type == 'subversion':
        return Subversion(config)
    elif res_type == 'git':
        return Git(config)
    elif res_type == 'included':
        return Resource(config)
    else:
        raise ConfigException('Uknown resource type.', config)
    
def instance_rosinstall_entry(entry, base_dir, errors=[]):
    if 'git' in entry:
        # print entry
        url = entry['git']['uri']
        dest = entry['git']['local-name']
        dest = os.path.join(base_dir, dest)
        branch = entry['git'].get('version', 'master')
        return Git(dict(dir=dest, url=url, branch=branch))
    elif 'tar' in entry:
        msg = 'skipping tar entry %r' % entry
        errors.append(Exception(msg)) 
        return None
    else:
        msg = 'Cannot interpret entry %r' % entry
        raise Exception(msg)

def find_configuration_in_dir(dirname, names=['resources.yaml', '.rosinstall']):
    for name in names:
        config = os.path.join(dirname, name)
     
        if os.path.exists(config):
            yield config

    
def find_configuration(dirname=os.path.curdir):
    while True:
        dirname = os.path.realpath(dirname)
        
        files = list(find_configuration_in_dir(dirname))
        
        for f in files:
            yield f
            
        if files:
            return
        
        parent = os.path.dirname(dirname)
        
        if parent == dirname:  # reached /
            raise Exception('Could not find configuration files.')
        
        dirname = parent
        

def load_resources(filename, errors=[]):
    basename = os.path.basename(filename)
    if 'resources.yaml' in basename:
        # print('loading %r' % filename)
        for x in load_resources_patience(filename, errors):
            yield x
        return
    
    if '.rosinstall' in basename:
        for x in load_resources_rosinstall(filename, errors):
            yield x
        return
    
    msg = 'Strange basename %r' % basename
    assert False, msg
    
def load_resources_rosinstall(filename, errors=[]):
    entries = list(yaml.load_all(open(filename)))[0]
    for entry in entries:
        i = instance_rosinstall_entry(entry, os.path.dirname(filename),
                                      errors=errors)
        if i is not None:
            yield i
    
def load_resources_patience(filename, errors=[]):
    curdir = os.path.dirname(filename)
    
    contents = list(yaml.load_all(open(filename)))
    if isinstance(contents[0], list):
        contents = contents[0]
        
    for config in contents:
        if config is None:
            continue
        config['from'] = filename
        sub = config.get('sub', None)
        if sub:
            f = os.path.join(curdir, sub)
            
            if not os.path.exists(f):
                msg = ('Could not load sub %s.' % f)
                errors.append(Exception(msg))
                continue
            
            if os.path.isdir(f):
                files = find_configuration_in_dir(f)
            else:
                files = [f]
                
            try:
                for ff in files:
                    for x in load_resources(ff, errors):
                        yield x
                
            except Exception as e:
                msg = ('Could not load %r: %s' % (sub, e))
                errors.append(Exception(msg))
            
        else:
            yield instantiate(config, curdir)



from collections import namedtuple
   
        
status_fields = ('present num_modified num_untracked to_push simple_push '
                'to_merge simple_merge branch current_branch branch_mismatch '
                'local_branch_exists remote_branch_exists url current_url').split()
StatusResult = namedtuple('StatusResult', status_fields)

__all__ = ['StatusResult', 'status_fields', 'status2string']

sizes = [2, 14, 8, 8, 4, 4, 4]
flag_modified = 1
flag_merge = 2
flag_push = 3
flag_branch = 4
flag_branch_remote = 5
flag_branch_local = 6


def status2string(r, res):
    
    flags = [''] * len(sizes)
    
    
    remarks = []
    
    if not res.present:
        flags[flag_modified] = 'missing'
    else:    
        if res.num_modified or res.num_untracked:
            fm = '%3dm' % res.num_modified if res.num_modified else "    "
            if res.num_modified > 99:
                fm = '>99u'
            fu = '%3du' % res.num_untracked if res.num_untracked else "    "
            if res.num_untracked > 99:
                fu = '>99u'
        
            flags[flag_modified] = fm + ' ' + fu
        
        if  (res.current_url is not None) and (res.url != res.current_url):
            remarks.append('Wrong remote url: %s' % res.current_url)
            remarks.append('(expected %s)' % res.url)
            
        if  res.local_branch_exists == False:
            flags[flag_branch_local] = 'no local %s' % res.branch
            flags[flag_branch_local] = '!L'
            
            remarks.append('Local branch %r does not exist.' % res.branch)
        
        if  res.remote_branch_exists == False:
            flags[flag_branch_remote] = 'no remote %s' % res.branch
            flags[flag_branch_remote] = '!R' 
            
            remarks.append('Remote branch %r does not exist.' % res.branch)

        
        if res.branch_mismatch == True:
            flags[flag_branch] = 'on %s instead of %s' % (res.current_branch, res.branch)
            flags[flag_branch] = '~B'
            
            remarks.append('On branch %r rather than %r.' % (res.current_branch, res.branch))
        else:                
            if res.to_merge == True:
                flags[flag_merge] = 'M %d' % res.to_merge
                if res.simple_merge == False:
                    flags[flag_merge] += '!'
                    remarks.append('There might be conflicts when merging.')
                else:
                    pass
                    # flags[flag_merge] += '    '            
    
            if res.to_push == True:
                flags[flag_push] = 'P %d' % res.to_push
                if res.simple_push == False:
                    flags[flag_push] += '!'
                    remarks.append('There might be conflicts when pushing.')
                else:
                    pass
                    # flags[flag_push] += '    '
                         
            
    if not all([f == '' for f in flags]) or remarks:
        status = ""
        for i, f in enumerate(flags):
            fm = "{0:<%d}" % sizes[i]
            status += fm.format(f)
        status += " {0}".format(r)
        if remarks:
            lleft = sum(sizes)  # + (len(sizes) - 1)
            prefix = ' ' * lleft 
            for r in remarks:
                status += '\n' + prefix + '  - %s' % r
                
        return status
    else:
        return None





# import sys
# import subprocess
# 
# def cmd2args(s):
#     ''' if s is a list, leave it like that; otherwise split()'''
#     if isinstance(s, list):
#         return s
#     elif isinstance(s, str):
#         return s.split() 
#     else: 
#         assert False
# 
# 
# class CmdResult(object):
#     def __init__(self, cwd, cmd, ret, stdout, stderr):
#         self.cwd = cwd
#         self.cmd = cmd
#         self.ret = ret
#         self.stdout = stdout
#         self.stderr = stderr
#         
#     def format(self):
#         return result_format(self.cwd, self.cmd, self.ret, self.stdout, self.stderr)
#     
#     
# class CmdException(Exception):
#     def __init__(self, cmd_result):
#         Exception.__init__(self, cmd_result.format())
#         self.res = cmd_result
#         
# def system_cmd_result(
#     cwd, cmd,
#     display_stdout=False,
#     display_stderr=False,
#     raise_on_error=False,
#     display_prefix=None):
#     
#     ''' Returns a tuple CmdResult; raises CmdException. '''
#     if display_prefix is None:
#         display_prefix = '%s %s' % (cwd, cmd)
#     
#     p = subprocess.Popen(
#             cmd2args(cmd),
#             stdout=subprocess.PIPE,
#             stderr=subprocess.PIPE,
#             cwd=cwd)
#     
#     if 1:  # XXX?
#         stdout, stderr = p.communicate()
#         
#         stdout = stdout.strip()
#         stderr = stderr.strip()
#         
#         prefix = display_prefix + 'err> '
#         if display_stderr and stderr:
#             print(indent(stderr, prefix))
#             
#         prefix = display_prefix + 'out> '
#         if display_stdout and stdout:
#             print(indent(stdout, prefix))
#     
#         p.wait()
#     else:
#         # p.stdin.close()
#         stderr = ''
#         stdout = ''
#         stderr_lines = []
#         stdout_lines = []
#         stderr_to_read = True
#         stdout_to_read = True
#         
#         def read_stream(stream, lines):
#             if stream:
#                 nexti = stream.readline()
#                 if not nexti:
#                     stream.close()
#                     return False
#                 lines.append(nexti)
#                 return True
#             else:
#                 stream.close()
#                 return False
#                 
#         # XXX: read all the lines
#         while stderr_to_read or stdout_to_read:
#             
#             if stderr_to_read:
#                 stderr_to_read = read_stream(p.stderr, stderr_lines)
# #             stdout_to_read = False
#         
#             if stdout_to_read:
#                 stdout_to_read = read_stream(p.stdout, stdout_lines)
#             
#             while stderr_lines:
#                 l = stderr_lines.pop(0)
#                 stderr += l
#                 if display_stderr:
#                     sys.stderr.write('%s ! %s' % (display_prefix, l))
#                     
#             while stdout_lines:
#                 l = stdout_lines.pop(0)
#                 stdout += l
#                 if display_stdout:
#                     sys.stderr.write('%s   %s' % (display_prefix, l))
#                 
#         stdout = p.stdout.read()
#         p.wait()
#             
#     ret = p.returncode 
#     
#     res = CmdResult(cwd, cmd, ret, stdout, stderr)
#     
#     if raise_on_error:
#         if res.ret != 0:
#             raise CmdException(res)
#     
#     return res
#         
# 
# def system_cmd_show(cwd, cmd): 
#     ''' Display command, raise exception. '''
#     system_cmd_result(
#             cwd, cmd,
#             display_stdout=True,
#             display_stderr=True,
#             raise_on_error=True)
#         
# def system_cmd(cwd, cmd):
#     ''' Do not output; return value. '''
#     res = system_cmd_result(
#             cwd, cmd,
#             display_stdout=False,
#             display_stderr=False,
#             raise_on_error=False)
#     return res.ret
# 
# def system_run(cwd, cmd):
#     ''' Gets the output of a command,  raise exception if it failes '''
#     res = system_cmd_result(
#             cwd, cmd,
#             display_stdout=False,
#             display_stderr=False,
#             raise_on_error=True)
#     return res.stdout
# 
# # todo: remove these
# system_cmd_fail = system_run
# system_output = system_run
#     
# def wrap(header, s, N=30):
#     header = '  ' + header + '  '
#     l1 = '-' * N + header + '-' * N
#     l2 = '-' * N + '-' * len(header) + '-' * N
#     return  l1 + '\n' + s + '\n' + l2
# 
# def result_format(cwd, cmd, ret, stdout=None, stderr=None):
#     msg = ('Command:\n\t{cmd}\n'
#            'in directory:\n\t{cwd}\nfailed with error {ret}').format(
#             cwd=cwd, cmd=cmd, ret=ret
#            )
#     if stdout is not None:
#         msg += '\n' + wrap('stdout', stdout)
#     if stderr is not None:
#         msg += '\n' + wrap('stderr', stderr)
#     return msg
#     
# 
# def indent(s, prefix):
#     lines = s.split('\n')
#     lines = ['%s%s' % (prefix, line.rstrip()) for line in lines]
#     return '\n'.join(lines)



from patience.action import Action
from patience.structures import ActionException

__all__ = ['Merge']



class Merge(Action):
    
    def __init__(self):
        Action.__init__(self, parallel=True, any_order=True)

    def single_action_started(self, resource): 
        pass
    
    def single_action_result_display(self, resource, result): 
        pass

    def single_action(self, r):
        if not r.is_downloaded():
            raise ActionException('Not downloaded %s' % r)
            
        # TODO: add branch check
        
        if r.something_to_merge() and r.simple_merge():
            r.merge()
                
Action.actions['merge'] = Merge()



from .resources import Resource
from system_cmd import  system_cmd_result, system_cmd_show, CmdException
from contracts import contract
from .structures import ActionException

__all__ = ['Git']


class Git(Resource):
    def __init__(self, config):
        Resource.__init__(self, config)
        self.url = config['url']
        self.branch = config.get('branch', 'master')
        
        self.show_operations = False 
        self.show_stdout = False
        
    def current_branch(self):
        branch = self.run('git rev-parse --abbrev-ref HEAD').strip()
        
        return branch + ''

    def check_right_branch(self):
        """ Raise exception if we are not on the right branch. """
        if not self.is_right_branch():
            cur_branch = self.current_branch()
            msg = ('Currently on branch %r rather than %r' % 
                    (cur_branch, self.branch)) 
            raise ActionException(msg)
    
    def checkout_right_branch(self):
        return self.runf('git checkout {branch}') 
    
    def is_right_branch(self):
        """ return false  if we are not on the branche
            declared in the configuration. """
        cur_branch = self.current_branch()
        return cur_branch == self.branch
            
    def check_remote_correct(self):
        if not self.is_remote_correct():
            msg = 'The remote is not correct: %s' % self.get_remote_url()
            raise ActionException(msg)
                
    def is_remote_correct(self):
        return self.get_remote_url() == self.url
    
    def get_remote_url(self):
        url = self.run('git config --get remote.origin.url')
        return url
        
    def check_branch_exists_remote(self):
        self.check_remote_correct()
        if not self.branch_exists_remote():
            msg = 'Remote branch %r does not exist.' % self.branch
            raise ActionException(msg)
            
    def branch_exists_remote(self, branch=None):
        if branch is None:
            branch = self.branch
        res = self.run0('git show-ref --verify refs/remotes/origin/%s' % branch,
                        raise_on_error=False)
        exists = res.ret == 0
        return exists

    def branch_exists_local(self):
        res = self.run0('git show-ref --quiet --verify -- refs/heads/%s' % self.branch,
                        raise_on_error=False)
        exists = res.ret == 0
        return exists
    
    def make_local_branch(self):
        return self.runf('git branch {branch}')

    def checkout(self):
        self.run(['git', 'clone', self.url, self.destination],
                    cwd='.'  # the other was not created yet
                )

    def run(self, cmd, cwd=None, errmsg=None): 
        return self.run0(cmd, cwd, errmsg).stdout
    
    def run0(self, cmd, cwd=None, errmsg=None, raise_on_error=True): 
        try:
            if cwd is None: 
                cwd = self.destination
                
            display_prefix = '%-30s: ' % self.short_path
            
            if self.show_operations:
                print(display_prefix + ' %s ' % cmd)
               
            display_prefix = '%-30s  ' % ''
            res = system_cmd_result(cwd, cmd,
                                    raise_on_error=raise_on_error,
                                    display_stdout=self.show_operations,
                                    display_stderr=self.show_operations,
                                    display_prefix=display_prefix
                                    )
            return res
        except CmdException as e:
            if errmsg:
                s = '%s\n' % errmsg
            else:
                s = 'Command %r failed with error code %d.' % (cmd, e.res.ret)
                if e.res.stdout:
                    s += ' stdout: %r ' % e.res.stdout
                if e.res.stderr:
                    s += ' stderr: %r ' % e.res.stderr          
                
            raise ActionException(s)
            
    def f(self, f, **args):
        ''' formats a string '''
        return f.format(branch=self.branch, **args)

    def runf(self, f, **args):
        ''' Formats and runs a command. '''
        return self.run(self.f(f, **args))
        
    def badconf(self, e):
        raise ActionException(e)

    def fetch(self):
        self.check_remote_correct()
        self.run('git fetch')
            
    def update(self):
        self.fetch()
        self.merge()

    def merge(self):
        self.check_remote_correct()
        self.check_right_branch()
        self.check_branch_exists_remote()
        if self.something_to_merge():
            if self.simple_merge():
                self.runf('git merge origin/{branch}')
            else:
                self.badcond("Cannot merge, because more than a FF is required.")
    

    def num_modified(self):
        output = self.run('git ls-files -m -d')
        files = linesplit(output)
        return len(files)

    def num_untracked(self):
        command = 'git ls-files --others --exclude-standard'  # ' --directory'
        output = self.run(command)
        files = linesplit(output)
        return len(files)

    @contract(returns='dict(str:str)')
    def list_remote_branches(self, ignore=['gh-pages']):
        """ REturns dictionary branch -> sha """
        out = self.run('git ls-remote --heads origin')
        # f93ffca589ad79a3ad2aad32019bf608d43b0956    refs/heads/env_dvsd
        # b4e8e4a0ec8432de4b6c1c61280199f0f3a270bf    refs/heads/env_fault
        # b4e8e4a0ec8432de4b6c1c61280199f0f3a270bf    refs/heads/master
        branches = {}
        for line in linesplit(out):
            tokens = line.split('\t')
            assert len(tokens) == 2, tokens
            sha = tokens[0]
            ref = tokens[1]
            assert 'refs/heads/' in ref
            branch = ref.replace('refs/heads/', '')
            assert branch
            if branch in ignore:
                continue
            branches[branch] = sha
        
        return branches
    
    @contract(returns='dict(str:tuple(int, bool, int, bool))')
    def list_differences_with_remote_branches(self):
        """
            Compares the current branch to the others in the remote.
            Returns dict rbranch -> npush, simple, nmerge, simple_merge.
        """
        res = {}
        for rbranch in self.list_remote_branches():
            lbranch = self.branch
            if lbranch == rbranch:
                continue
            npush = self.something_to_push_to(lbranch=lbranch, rbranch=rbranch)
            if npush > 0:
                simple_push = self.simple_push_generic(lbranch=lbranch, rbranch=rbranch)
            else:
                simple_push = True
            nmerge = self.something_to_merge_from(lbranch=lbranch, rbranch=rbranch)
            if nmerge > 0:
                simple_merge = self.simple_merge_generic(lbranch=lbranch, rbranch=rbranch)
            else:
                simple_merge = True
            res[rbranch] = (npush, simple_push, nmerge, simple_merge) 
        return res
     
    def something_to_push_to(self, lbranch, rbranch):
        ''' Returns the number of commits that we can push to a remote
            branch.'''
        self.check_right_branch()
        self.check_remote_correct()
        if not self.branch_exists_remote(rbranch):
            raise ValueError(rbranch)
        
        output = self.runf('git log origin/{rbranch}..{lbranch} --no-merges --pretty=oneline',
                           rbranch=rbranch, lbranch=lbranch)
        commits = linesplit(output)
        return len(commits)
    
    def something_to_merge_from(self, lbranch, rbranch):
        ''' Returns the number of commits that we can merge from a remote branch.'''
        self.check_right_branch()
        self.check_remote_correct()
        # if not self.branch_exists_remote():
        #    return 0
        if not self.branch_exists_remote(rbranch):
            raise ValueError(rbranch)
        output = self.runf('git log  {lbranch}..origin/{rbranch} --no-merges --pretty=oneline',
                           rbranch=rbranch, lbranch=lbranch)
        commits = linesplit(output)
        return len(commits)
    
    def simple_merge_generic(self, lbranch, rbranch):
        ''' Checks that the local lbranch can be fast forwarded to rbranch. '''
        self.check_remote_correct()
#         self.check_branch_exists_remote(rbranch)
        rev = self.runf('git rev-parse {lbranch}', lbranch=lbranch).strip()
        base = self.runf('git merge-base {rev} origin/{rbranch}', rev=rev, rbranch=rbranch)
        if rev == base.strip():
            return True
        else:
            return False
    
    def simple_push_generic(self, lbranch, rbranch):
        ''' Returns true if we can do a safe push (assuming we have the last
            revision of the remote branch.) '''
        self.check_remote_correct()
#         if not self.branch_exists_remote():
#             return True

        stdout = self.runf('git rev-list {lbranch}..origin/{rbranch}',
                           lbranch=lbranch, rbranch=rbranch)
        if stdout.strip():
            return False
        else:
            return True
        
    def something_to_push(self):
        ''' Returns the number of commits that we can push to the remote
            branch.'''
        self.check_right_branch()
        self.check_remote_correct()
        if not self.branch_exists_remote():
            # print('need to push brnach')
            # we need to push the branch, at least
            return 1 
        
        command = 'git log origin/{branch}..{branch} --no-merges --pretty=oneline'
        output = self.runf(command)
        commits = linesplit(output)
        return len(commits)
    
    def something_to_merge(self):
        ''' Returns the number of commits that we can merge from remote branch.'''
        self.check_right_branch()
        self.check_remote_correct()
        if not self.branch_exists_remote():
            return 0
        
        command = 'git log {branch}..origin/{branch} --no-merges --pretty=oneline'
        output = self.runf(command)
        commits = linesplit(output)
        return len(commits)
    
    def simple_merge(self):
        ''' Checks that our branch can be fast forwarded. '''
        self.check_right_branch() 
        self.check_remote_correct()
        self.check_branch_exists_remote()
        rev = self.runf('git rev-parse {branch}').strip()
        base = self.runf('git merge-base {rev} origin/{branch}', rev=rev)
        if rev == base.strip():
            return True
        else:
            return False
    
    def simple_push(self):
        ''' Returns true if we can do a safe push (assuming we have the last
            revision of the remote branch.) '''
        self.check_right_branch()
        self.check_remote_correct()
        if not self.branch_exists_remote():
            return True

        stdout = self.runf('git rev-list {branch}..origin/{branch}')
        if stdout.strip():
            return False
        else:
            return True
    
    def commit(self):
        self.check_right_branch()
        if self.num_modified():
            n = self.num_untracked()
            if n > 0:
                self.badcond("Cannot commit; there are %s untracked files." % n)
            else: 
                # TODO
                system_cmd_show(self.destination, 'git status')
                try:
                    msg = raw_input('message: ')
                except Exception as e:
                    print e
                    return
                self.run(self.destination, ['git', 'commit', '-a', '-m', msg])
    
    def push(self):
        self.check_right_branch()
        self.check_remote_correct()

        if not self.simple_push():
            self.badcond("Cannot push because of conflicts.")
            
        self.runf('git push origin {branch}')
            

    def current_revision(self):
        out = self.run('git rev-parse HEAD')
        out = out.split()[0]
        return out

def linesplit(s):
    ''' Splits a string in lines; removes empty. '''
    return filter(lambda x: x, s.split('\n'))
    



from patience.action import Action
from patience.status_string import status2string, StatusResult, status_fields


__all__ = ['Status', 'StatusFull']


class Status(Action):
    
    def __init__(self): 
        Action.__init__(self, parallel=True, any_order=False)

    def single_action_starting(self, resource):  # @UnusedVariable
        return None

    def result_display_start(self):
        return """  unknown_       __merges    
modified  |     |    pushes     _branch status
    |     |     |       |      |
"""

    def single_action_result_display(self, resource, result):
        if not isinstance(result, Exception):
            return status2string(resource, result)
    
    def single_action(self, r):
        
        branch = r.branch
        url = r.url

        current_url = None
        num_modified = None
        num_untracked = None
        to_push = None
        to_merge = None
        simple_merge = None
        simple_push = None
        current_branch = None
        branch_mismatch = None
        local_branch_exists = None
        remote_branch_exists = None

        if not r.is_downloaded():
            present = False
        else:
            present = True
            current_url = r.get_remote_url()
            
            if current_url != url:
                pass
            else:
                remote_branch_exists = r.branch_exists_remote()
               
                current_branch = r.current_branch()
                branch_mismatch = branch != current_branch
                local_branch_exists = r.branch_exists_local()
                
                if branch == current_branch:
                    to_push = r.something_to_push()
                    to_merge = r.something_to_merge()
                    if to_merge: 
                        simple_merge = r.simple_merge()
                    else: 
                        simple_merge = None
                    if to_push: 
                        simple_push = r.simple_push()
                    else: 
                        simple_push = None
                else:
                    to_push = None
                    to_merge = None
                    simple_merge = None
                    simple_push = None
                

            num_modified = r.num_modified()
            num_untracked = r.num_untracked()
            
        asdict = dict([(k, locals()[k]) for k in status_fields])
        return StatusResult(**asdict)

    def summary(self, resources, results):
        pass

Action.actions['status'] = Status()


class StatusFull(Status):
    

    def single_action_result_display(self, resource, result):
        if not isinstance(result, Exception):
            return status2string(resource, result)
        else:
            return str(result)  # XXX


Action.actions['status-full'] = StatusFull()




from .structures import UserError
import datetime
import platform
import sys


def main():
    from optparse import OptionParser
    parser = OptionParser()
    parser.add_option("--config", help="Location of yaml configuration")

    parser.add_option("-s", "--seq", help="Force sequential", default=False,
                    action='store_true')

    parser.add_option("-v", "--verbose", help="Write status messages",
                    default=False, action='store_true')
                
    parser.add_option("-V", help="Show git operations",
                     dest='show_operations',
                    default=False, action='store_true')
    
    parser.add_option("-O", help="Show stdout/stderr of operations",
                     dest='show_stdout',
                    default=False, action='store_true')
    
    
    parser.add_option("--yaml", help="Write YAML output", default=False,
                    action='store_true')

    (options, args) = parser.parse_args()  # @UnusedVariable

    from .configuration import find_configuration, load_resources

    if options.config:
        configs = [options.config]
    else:    
        configs = list(find_configuration())
        
    resources = []
    errors = []
    found = {}
    for c in configs:
        for r in load_resources(c, errors):
            if r.destination in found:
                continue
            else:
                found[r.destination] = r 
                resources.append(r) 
         
    
    for e in errors:
        print('error: %s' % e)
        
    from .action import Action
    if len(args) == 0:
        msg = 'Please provide command: %s' % (Action.actions.keys()) 
        raise UserError(msg)
    
    if len(args) > 1:
        raise Exception('Please provide only one command.')
    command = args[0]
    
    
    quiet = False


    if options.yaml:
        stream = None
    else:
        stream = sys.stdout
        
    
    if command in Action.actions:
        action = Action.actions[command]
        
        force_sequential = options.seq
        if command in ['status'] and options.yaml:
            force_sequential = True
            
        results = action.go(resources,
            force_sequential=force_sequential,
            stream=stream,
            console_status=options.verbose,
            show_operations=options.show_operations,
            show_stdout=options.show_stdout)
        
        if options.yaml:
            import yaml
            s = {'date': datetime.datetime.now(),
                 'hostname': platform.node(),
                 'command': command,
                 'config': configs,
                 'resources': resources,
                 'results': results}
            # yaml.safe_dump(s, sys.stdout, default_flow_style=False)
            yaml.dump(s, sys.stdout, default_flow_style=False)
        return
        
    if command == 'list':
        repos = [dict(dir=r.destination, url=r.url)  for r in resources]
        yaml.dump(repos, sys.stdout, default_flow_style=False) 
                     
    elif command == 'update':
        for r in resources:
            if not quiet:
                print('Updating %s' % r)

            r.update()

    elif command == 'tag':
        h = []
        for r in resources:
            c = r.config.copy()
            c['revision'] = r.current_revision()
            h.append(c)
        print(yaml.dump(h))
         
    elif command == 'commit':
        for r in resources:
            if r.num_modified() > 0 and  r.num_untracked() == 0:
                r.commit()
    else:
        raise Exception('Unknown command "%s".' % command)
        # 
        # def fetch(r):
        #     if r.config['type'] == 'git':
        #         print 'Fetching for %s' % r
        #         res = r.fetch()
        #         if res:  
        #             print "fetched {dir}".format(dir=r)
        

if __name__ == '__main__':
    main()
       




import yaml
import sys

def main():
    
    data = yaml.load(sys.stdin)
    
    f = sys.stdout
    
    f.write("""
    
<html><head><title>Patience results</title>
    
    <style type="text/css">
    body { font-family: Verdana; } 
    tr.clean { display: none; }

    p, tr.header { font-style: italic; font-family: serif; }
    
    tr { padding: 0; margin: 0; border: 0; }
    tr.header{ border-bottom: solid 1px gray; }
    tr.header td { text-align: center }
/*  td.name { border-right: solid 2px gray; }*/
    td { padding: 0; margin: 0; border: 0; 
        }
    
    span.zero { display: none }
    td.name { font-weight: bold; text-align: left; }
    td.number {  width: 3em; font-family: monospace; text-align: center;}
    
    tr.some_modified .num_modified { background-color: orange; }
    tr.missing  { background-color: red; }
    /* tr.some_untracked .num_untracked { background-color: red; } */
    tr.some_to_push .to_push { background-color: yellow; }
    tr.some_to_merge .to_merge { background-color: #aaf; }
    /* TODO: add simple_merge/push */
    
    </style>

</head>
<body>
    
""")
    
    assert data['command'] == 'status'
    
    d = str(data['date'])[:16]
    ts = """ %s @ %s """ % (data['hostname'], d)
    
    f.write("""<table>
    
        <tr class='header'>
            <td> %s</td>
            <td> M </td>
            <td> ? </td>
            <td> push </td> 
            <td> merge </td>
            <td> notes </td>
        </tr>
    """ % ts)
    for r in data['resources']:
        # if not isinstance(status, patience.action.StatusResult):


        short_path = r.short_path
        status = data['results'][short_path]
        t = """
        <tr class="{tr_class}">
            <td class="name">{name}</td>
            <td class="number num_modified">{num_modified}</td>
            <td class="number num_untracked">{num_untracked}</td>
            <td class="number to_push">{to_push}</td>
            <td class="number to_merge">{to_merge}</td>
            <td class="notes">{notes}</td>
        </tr>
        """

        def wrap(x):
            if x:
                return "<span class='some'>%s</span>" % x
            else:
                return "<span class='zero'>%s</span>" % x
    
        classes = ["resource"]
        
        if isinstance(status, Exception):
            classes.append('exception') 
            invalid = '-'
            s = t.format(name=short_path,
                        num_modified=invalid,
                        num_untracked=invalid,
                        to_push=invalid,
                        to_merge=invalid,
                        tr_class=" ".join(classes),
                        notes='%s' % status)

        else:
            if not status.present:
                classes.append('missing')
            else:
                if status.num_modified:
                    classes.append('some_modified')
                else: 
                    classes.append('not_modified')
            
                if status.num_untracked: 
                    classes.append('some_untracked')
                else: 
                    classes.append('not_untracked')
                
                if status.to_merge: 
                    classes.append('some_to_merge')
                else: 
                    classes.append('not_to_merge')

                if status.to_push: 
                    classes.append('some_to_push')
                else: 
                    classes.append('not_to_push')
            
                if not (status.num_modified or 
                    status.num_untracked or
                    status.to_merge or 
                    status.to_push): 
                    classes.append('clean')

                
            s = t.format(name=short_path,
                        num_modified=wrap(status.num_modified),
                        num_untracked=wrap(status.num_untracked),
                        to_push=wrap(status.to_push),
                        to_merge=wrap(status.to_merge),
                        tr_class=" ".join(classes),
                        notes="")
        
        f.write(s)
        
    f.write('</table>\n')

    f.write("""</body></html>""")
    
if __name__ == '__main__':
    main()



from patience.action import Action
from patience.structures import ActionException

__all__ = ['MakeBranch']


class MakeBranch(Action):
    """ Makes the required local branch if it does not exist and switches to it. """
    
    def single_action_starting(self, resource): 
        pass

    def single_action_result_display(self, resource, result): 
        pass

    def applicable(self, r):
        return not r.is_right_branch()
        
    def single_action(self, r):
        if r.is_right_branch():
            raise ActionException('Already on branch')
        
        if not r.branch_exists_local():
            r.make_local_branch()
            
        if not r.is_right_branch():
            r.checkout_right_branch() 


    
Action.actions['make-branch'] = MakeBranch()




from setuptools import setup, find_packages


setup(name='patience',
      author="Andrea Censi",
      author_email="andrea@cds.caltech.edu",
      url='http://github.com/AndreaCensi/patience',
    	version="1.2.1",
      description="A command line tool for managing multiple git repositories.",
      package_dir={'':'src'},
      packages=find_packages('src'),
      install_requires=['pyyaml', 'SystemCmd', 'PyContracts'],
      entry_points={
         'console_scripts': [
           'patience = patience.main:main',
           'patience_search = patience.patience_search:main',
           'pat = patience.main:main',
           'patience2html = patience.patience2html:main'
        ]  
      },
)





from patience.action import Action
from patience.structures import ActionException

__all__ = ['Fetch']

        
class Fetch(Action):
    
    def __init__(self):
        Action.__init__(self, parallel=True, any_order=True)

    def single_action_result_display(self, resource, result):  # @UnusedVariable
        if isinstance(result, Exception):
            return "%s" % result
        else:
            if result:
                return "fetched: %s" % result
            else:
                return None
            
    def single_action(self, r):
        if not r.is_downloaded():
            raise ActionException('Not downloaded %s' % r)

#         if isinstance(r, Git):
        return r.fetch()
        
#         raise ActionException("Not implemented for %r" % r.config['type'])

Action.actions['fetch'] = Fetch()

        